# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .action_result_data import ActionResultData
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class ActionResult(UniversalBaseModel):
    success: bool
    stop_execution_on_failure: typing.Optional[bool] = None
    exception_type: typing.Optional[str] = None
    exception_message: typing.Optional[str] = None
    data: typing.Optional[ActionResultData] = None
    step_retry_number: typing.Optional[int] = None
    step_order: typing.Optional[int] = None
    download_triggered: typing.Optional[bool] = None
    interacted_with_sibling: typing.Optional[bool] = None
    interacted_with_parent: typing.Optional[bool] = None
    skip_remaining_actions: typing.Optional[bool] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
