# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .action_result import ActionResult
from .user_defined_error import UserDefinedError
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class AgentStepOutput(UniversalBaseModel):
    """
    Output of the agent step, this is recorded in the database.
    """

    action_results: typing.Optional[typing.List[ActionResult]] = None
    actions_and_results: typing.Optional[typing.List[typing.List[typing.Optional[typing.Any]]]] = None
    errors: typing.Optional[typing.List[UserDefinedError]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
