# This file was auto-generated by Fern from our API Definition.

import typing

EntityType = typing.Union[
    typing.Literal["step", "task", "workflow_run", "workflow_run_block", "observer_thought"], typing.Any
]
