# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import datetime as dt
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class TaskGeneration(UniversalBaseModel):
    organization_id: str
    user_prompt: str
    user_prompt_hash: str
    url: typing.Optional[str] = None
    navigation_goal: typing.Optional[str] = None
    navigation_payload: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    data_extraction_goal: typing.Optional[str] = None
    extracted_information_schema: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    llm: typing.Optional[str] = None
    llm_prompt: typing.Optional[str] = None
    llm_response: typing.Optional[str] = None
    suggested_title: typing.Optional[str] = None
    task_generation_id: str
    created_at: dt.datetime
    modified_at: dt.datetime

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
